from blatann.uuid import Uuid16

CURRENT_TIME_SERVICE_UUID = Uuid16(0x1805)

CURRENT_TIME_CHARACTERISTIC_UUID = Uuid16(0x2A2B)
LOCAL_TIME_INFO_CHARACTERISTIC_UUID = Uuid16(0x2A0F)
REFERENCE_INFO_CHARACTERISTIC_UUID = Uuid16(0x2A14)
