from blatann.uuid import Uuid16


GLUCOSE_SERVICE_UUID = Uuid16(0x1808)

MEASUREMENT_CHARACTERISTIC_UUID = Uuid16(0x2A18)
MEASUREMENT_CONTEXT_CHARACTERISTIC_UUID = Uuid16(0x2A34)
FEATURE_CHARACTERISTIC_UUID = Uuid16(0x2A51)
RACP_CHARACTERISTIC_UUID = Uuid16(0x2A52)

